from __future__ import absolute_import, print_function, division

from pyramid_frontend.theme import Theme
from pyramid_frontend.images.chain import FilterChain
from pyramid_frontend.assets.less import LessAsset


def base_includeme(config):
    test_includes = config.registry.settings.setdefault('test_includes', [])
    test_includes.append('base')


class BaseTheme(Theme):
    key = 'base'
    image_filters = [
        FilterChain('full', width=400, height=400, crop=True),
    ]
    assets = {
        'main-less': LessAsset('/_base/css/main.less'),
    }
    includes = [
        base_includeme,
    ]



from __future__ import absolute_import, print_function, division

import logging
import time
import subprocess

log = logging.getLogger(__name__)


def run(args):
    log.debug('Running command: %s ...', ' '.join(args))
    start_time = time.time()
    try:
        subprocess.check_output(args, stderr=subprocess.STDOUT, close_fds=True)
    except subprocess.CalledProcessError as e:
        log.error(e.output)
        raise
    finally:
        elapsed_time = time.time() - start_time
        log.debug('Command completed in %0.4f seconds.', elapsed_time)



from __future__ import absolute_import, print_function, division

import re
from unittest import TestCase
from six import BytesIO

from webtest import TestApp

from PIL import Image

from ..images import files
from ..images.view import MissingOriginal
from ..templating.renderer import MakoRenderingException

from . import utils


def setup():
    utils.load_images()


class Functional(TestCase):
    settings = None

    def setUp(self):
        self.app = TestApp(utils.make_app(self.settings))


class TestTemplatingFunctional(Functional):
    def test_render_index(self):
        resp = self.app.get('/')
        resp.mustcontain('base theme base.html')
        resp.mustcontain('foo theme base.html')
        resp.mustcontain('base theme index.html')

    def test_render_bad_return(self):
        with self.assertRaises(ValueError):
            self.app.get('/bad-return')

    def test_render_bad_template(self):
        with self.assertRaises(MakoRenderingException) as e:
            self.app.get('/bad-template')
        self.assertIn('dummy error', repr(e.exception))

    def test_render_text_escaping(self):
        resp = self.app.get('/text-template')
        resp.mustcontain('should not be escaped: <foo>')
        resp.mustcontain('should be escaped: &lt;foo&gt;')


class TestThemeStrategy(Functional):
    def setUp(self):
        def theme_strategy(req):
            return req.params.get('theme', 'foo')
        self.app = TestApp(utils.make_app(theme_strategy=theme_strategy))

    def test_render_alternate_theme(self):
        resp = self.app.get('/?theme=base')
        self.assertIn(b'base', resp.body)
        self.assertNotIn(b'foo', resp.body)

        resp = self.app.get('/?theme=foo')
        self.assertIn(b'base', resp.body)
        self.assertIn(b'foo', resp.body)


class TestAssetsFunctional(Functional):
    def test_js_tag(self):
        resp = self.app.get('/js-tag')
        resp.mustcontain('main.js')
        resp.mustcontain('require.js')

    def test_css__tag(self):
        resp = self.app.get('/css-tag')
        resp.mustcontain('main.less')
        resp.mustcontain('less.js')

    def test_svg_tag(self):
        resp = self.app.get('/svg-tag')
        resp.mustcontain('logo.svg')
        resp.mustcontain('img')


class TestCompiledFunctional(Functional):
    settings = {
        'pyramid_frontend.compile': True,
    }

    def setUp(self):
        Functional.setUp(self)
        utils.compile_assets()

    def test_js_tag(self):
        resp = self.app.get('/js-tag')
        body = resp.body.decode('utf-8')
        self.assertNotIn('require.js', body)
        # Look for something that looks like a hex digest
        self.assertRegexpMatches(body, '[a-f0-9]{8,}')
        # Fetch a second time to test caching and make sure it's the same
        # result.
        resp2 = self.app.get('/js-tag')
        body2 = resp2.body.decode('utf-8')
        self.assertEqual(body, body2)

        # Extract the filepath and load it.
        m = re.search('src="([^\"]+)"', body)
        path = m.group(1)
        file_resp = self.app.get(path)
        self.assertGreater(len(file_resp.body), 0)

    def test_css_tag(self):
        resp = self.app.get('/css-tag')
        body = resp.body.decode('utf-8')
        self.assertNotIn('less.js', body)
        # Look for something that looks like a hex digest
        self.assertRegexpMatches(body, '[a-f0-9]{8,}')

        # Extract the filepath and load it.
        m = re.search('href="([^\"]+)"', body)
        path = m.group(1)
        file_resp = self.app.get(path)
        self.assertGreater(len(file_resp.body), 0)

    def test_svg_tag(self):
        resp = self.app.get('/svg-tag')
        resp.mustcontain('img')

        # FIXME Check for minification and stuff, loading file path.


class TestImagesFunctional(Functional):
    def setUp(self):
        self.app = TestApp(utils.make_app())

    def test_fetch_image(self):
        url_resp = self.app.get('/image-url')
        img_resp = self.app.get(url_resp.body.decode('utf-8'))
        # This image should be 200x200
        f = BytesIO(img_resp.body)
        im = Image.open(f)
        self.assertEqual(im.size, (200, 200))

    def test_qualified_image_url(self):
        url_resp = self.app.get('/image-url')
        qual_resp = self.app.get('/image-url?qualified=1')
        self.assertTrue(qual_resp.body.endswith(url_resp.body))
        self.assertTrue(qual_resp.body.startswith(b'http://'))

    def test_image_url_theme(self):
        self.app.get('/image-url?filter_key=full')

        with self.assertRaises(AssertionError):
            self.app.get('/image-url?filter_key=barthumb')

    def test_image_tag(self):
        url_resp = self.app.get('/image-url')
        tag_resp = self.app.get('/image-tag')
        tag_resp.mustcontain('200')
        tag_resp.mustcontain('height')
        tag_resp.mustcontain('width')
        tag_resp.mustcontain('width')
        tag_resp.mustcontain(url_resp.body)

    def test_image_original_path(self):
        resp = self.app.get('/image-original-path')
        resp.mustcontain('originals')

    def test_fetch_image_bad_prefix(self):
        self.app.get('/img/aaaa/smiley-jpeg-rgb_jpg_thumb.png', status=404)

    def test_fetch_image_bad_suffix(self):
        name = 'smiley-jpeg-rgb'
        prefix = files.prefix_for_name(name)
        self.app.get('/img/%s/%s_jpg_nonexistent.png' % (prefix, name),
                     status=404)

    def test_fetch_image_bad_extension(self):
        name = 'smiley-jpeg-rgb'
        prefix = files.prefix_for_name(name)
        self.app.get('/img/%s/%s_doc_thumb.png' % (prefix, name),
                     status=404)

    def test_fetch_twice(self):
        name = 'smiley-jpeg-rgb'
        prefix = files.prefix_for_name(name)
        resp_a = self.app.get('/img/%s/%s_jpg_thumb.png' % (prefix, name))
        # FIXME This should try to find a way to test that it's fetching from
        # the filesystem the second time.
        resp_b = self.app.get('/img/%s/%s_jpg_thumb.png' % (prefix, name))
        self.assertEqual(resp_a.body, resp_b.body)

    def test_fetch_missing_original(self):
        name = 'nonexistent-file'
        prefix = files.prefix_for_name(name)
        with self.assertRaises(MissingOriginal):
            self.app.get('/img/%s/%s_jpg_thumb.png' % (prefix, name))

    def test_fetch_malformed_url(self):
        name = 'smiley-jpeg-rgb'
        prefix = files.prefix_for_name(name)
        self.app.get('/img/%s/%s_thumb.png' % (prefix, name), status=404)


class TestImagesDebug(Functional):
    def setUp(self):
        settings = {
            'pyramid_frontend.debug': 'true',
        }
        self.app = TestApp(utils.make_app(settings))

    def test_fetch_missing_original_placeholder(self):
        name = 'nonexistent-file'
        prefix = files.prefix_for_name(name)
        img_resp = self.app.get('/img/%s/%s_jpg_thumb.png' % (prefix, name))
        f = BytesIO(img_resp.body)
        im = Image.open(f)
        self.assertEqual(im.size, (200, 200))



from __future__ import absolute_import, print_function, division

from webhelpers2.html.tags import literal
from pyramid.settings import asbool


def asset_tag(request, key, **kwargs):
    """
    Request method to render an HTML fragment containing tags which reference
    the supplied entry point. This will dispatch to the appropriate tag
    rendering function based on context and entry point type.
    """
    theme = request.theme
    asset = theme.stacked_assets[key]
    settings = request.registry.settings
    should_compile = asbool(settings.get('pyramid_frontend.compile'))

    if should_compile:
        filename = theme.compiled_asset_path(key)
        url_path = '/compiled/' + theme.key + '/' + filename
    else:
        url_path = asset.url_path

    return literal(asset.tag(theme, url_path, production=should_compile,
                             **kwargs))


def includeme(config):
    config.add_request_method(asset_tag, 'asset_tag')

    compiled_path = \
        config.registry.settings['pyramid_frontend.compiled_asset_dir']
    config.add_static_view(name='compiled', path=compiled_path)



from __future__ import absolute_import, print_function, division

from unittest import TestCase

from pyramid import testing
from pyramid.httpexceptions import HTTPNotFound

from ..images.view import ImageView


class TestImageView(TestCase):

    def test_view_notfound(self):
        with testing.testConfig():
            request = testing.DummyRequest()
            request.matchdict['prefix'] = 'aaaa'
            request.matchdict['name'] = 'nonexistent-image.jpg'
            with self.assertRaises(HTTPNotFound):
                ImageView(request)()



import sys
import os

try:
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
except ImportError:
    html_theme = 'default'

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.dirname(os.path.abspath('.')))

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.todo',
              'sphinx.ext.coverage',
              'sphinx.ext.viewcode',
              'sphinx.ext.graphviz']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'pyramid_frontend'
copyright = u'2012-2015, Scott Torborg'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '0.4'
# The full version, including alpha/beta/rc tags.
release = '0.6.4'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# -- Options for HTML output --------------------------------------------------

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = 'pyramid_frontend ' + version

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Output file base name for HTML help builder.
htmlhelp_basename = 'pyramid_frontenddoc'

# -- Options for LaTeX output -------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ('index', 'pyramid_frontend.tex', u'pyramid_frontend Documentation',
     u'Scott Torborg', 'manual'),
]

# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'pyramid_frontend', u'pyramid_frontend Documentation',
     [u'Scott Torborg'], 1)
]

# -- Options for Texinfo output -----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index',
     'pyramid_frontend',
     u'pyramid_frontend Documentation',
     u'Scott Torborg',
     'pyramid_frontend',
     'One line description of project.',
     'Miscellaneous'),
]



from __future__ import absolute_import, print_function, division

import shutil
import tempfile
import math
from six import BytesIO
from six.moves import xrange

from PIL import Image

from .. import cmd
from .utils import (pad_image, flatten_alpha, crop_entropy,
                    is_white_background, is_larger, bounding_box, sharpen)


class Filter(object):
    """
    Filter stage superclass. Instances are called with some input data and
    produce some output data. Has some useful methods for processing.

    Filter instances are called on input data and return output data. Filters
    can take an arbitrary input and output, but by convention will tend to pass
    either PIL images or file-like objects.
    """

    def adapt_input(self, input):
        """
        Given a PIL image or a file-like object, return the PIL image.
        """
        if hasattr(input, 'getpixel'):
            im = input
            assert im.mode != 'CMYK', \
                "CMYK-mode Image instance can't be passed directly to filter"
        else:
            input.seek(0)
            im = Image.open(input)
            # Handle CMYK images
            if im.mode == 'CMYK':
                pos = input.tell()
                input.seek(0)
                rgb_input = CMYKFilter()(input)
                input.seek(pos)
                im = Image.open(rgb_input)
        return im

    def __call__(self, input):
        return self.filter(self.adapt_input(input))

    def shell_process(self, input, args, inplace=False):
        """
        Process an image file (on the filesystem) using a shell command.
        """
        input.seek(0)
        temp = tempfile.NamedTemporaryFile()
        shutil.copyfileobj(input, temp)
        temp.flush()

        if inplace:
            out = temp
        else:
            out = tempfile.NamedTemporaryFile()

        processed_args = []
        for arg in args:
            if arg == 'IN':
                arg = temp.name
            elif arg == 'OUT':
                arg = out.name
            processed_args.append(arg)

        cmd.run(processed_args)
        out.seek(0)
        return out


class ThumbFilter(Filter):
    """
    A filter that resizes to a given size, using various mechanisms for
    changing size.
    """
    def __init__(self, dimensions, pad=False, crop=False,
                 crop_whitespace=False, crop_whitespace_pad=0,
                 background='white', enlarge=False):

        self.dimensions = dimensions
        self.pad = pad
        self.crop = crop
        self.crop_whitespace = crop_whitespace
        self.crop_whitespace_pad = crop_whitespace_pad
        self.background = background
        self.enlarge = enlarge

    def filter(self, im):
        def _pad_dim(src, dst, flag):
            assert isinstance(flag, bool)
            return (dst if flag else src)

        im = flatten_alpha(im, self.background)

        if self.crop is True:
            should_entropy_crop = True
        elif self.crop == 'nonwhite':
            should_entropy_crop = not is_white_background(im)
        else:
            should_entropy_crop = False

        bgcolor = im.convert('RGBA').getpixel((0, 0))

        # FIXME The cropping behavior is not really correct here for dimensions
        # that are partially unspecified.
        if should_entropy_crop and is_larger(im, self.dimensions):
            im = crop_entropy(im, self.dimensions)

        if self.crop_whitespace and is_larger(im, self.dimensions):
            has_white_background = is_white_background(im, tolerance=0)
            im = im.crop(bounding_box(im))
            if has_white_background:
                xpad = int(self.crop_whitespace_pad * im.size[0])
                ypad = int(self.crop_whitespace_pad * im.size[1])
                pad_dims = (im.size[0] + (2 * xpad),
                            im.size[1] + (2 * ypad))
                im = pad_image(im, pad_dims)

        # FIXME The enlarge flag should only have any effect if the image
        # actually needs to be enlarged.
        if self.enlarge:
            factor = max(float(self.dimensions[0]) / im.size[0],
                         float(self.dimensions[1]) / im.size[1])
            im = im.resize((int(math.ceil(im.size[0] * factor)),
                            int(math.ceil(im.size[1] * factor))),
                           Image.BICUBIC)

        w, h = im.size
        aspect = float(w) / float(h)
        desired_w, desired_h = self.dimensions
        if (not desired_w) and (not desired_h):
            desired_w, desired_h = w, h
        elif not desired_w:
            desired_w = aspect * desired_h
        elif not desired_h:
            desired_h = desired_w / aspect

        im.thumbnail((desired_w, desired_h), Image.ANTIALIAS)
        if self.pad:
            w = _pad_dim(im.size[0], self.dimensions[0], self.pad)
            h = _pad_dim(im.size[1], self.dimensions[1], self.pad)
            im = pad_image(im, (w, h), color=bgcolor)
        return im


class VignetteFilter(Filter):
    """
    A filter which vignettes corners a bit, to make a white background image
    stand out a bit against a white page background.
    """
    def __init__(self, falloff=4, extent=40):
        self.falloff = falloff
        self.extent = extent

    def filter(self, im):
        falloff = self.falloff
        extent = self.extent

        def length(start, end):
            start_x, start_y = start
            end_x, end_y = end
            dist_x = end_x - start_x
            dist_y = end_y - start_y
            return math.sqrt((dist_x ** 2) + (dist_y ** 2))

        def light_falloff(radius, outside):
            return ((radius / outside) ** falloff) * extent

        im = im.convert('RGBA')

        w, h = im.size
        center = w / 2, h / 2
        outside = length(center, (0, 0))

        data = []
        for y in xrange(h):
            for x in xrange(w):
                radius = length(center, (x, y))
                factor = light_falloff(radius, outside)
                data.append(factor)

        alpha_im = Image.new('L', im.size)
        alpha_im.putdata(data)
        overlay_im = Image.new('L', im.size, 'black')
        return Image.composite(overlay_im, im, alpha_im)


class CMYKFilter(Filter):
    """
    Convert a file-like object to RGB colorspace using ImageMagick. This should
    be quite robust to weird things like CMYK TIFFs.
    """

    def __call__(self, input):
        return self.shell_process(input, ['convert', 'IN',
                                          '-colorspace', 'rgb',
                                          'OUT'])


class JPGSaver(Filter):
    """
    A JPG saver. Accepts keyword arguments, which will be passed to PIL's
    ``save()`` method. Calls jpegoptim on output.
    """
    def __init__(self, sharpness=None, **kwargs):
        self.sharpness = sharpness
        self.kwargs = kwargs

    def filter(self, im):
        if self.sharpness:
            im = sharpen(im, self.sharpness)
        if im.mode.endswith("A"):
            ni = Image.new("RGB", im.size, 'white')
            ni.paste(im, im)
            im = ni
        elif im.mode != "RGB":
            im = im.convert('RGB')
        buf = BytesIO()
        im.save(buf, 'JPEG', **self.kwargs)
        buf.seek(0)
        return buf


class JPGProcessor(Filter):
    """
    Postprocess a JPEG. For now, just uses jpegoptim to do some additional
    lossless slimming.
    """
    def __call__(self, input):
        return self.shell_process(input,
                                  ['jpegoptim', '--strip-all', 'IN'],
                                  inplace=True)


class PNGSaver(Filter):
    """
    Save a file as a 24-bit PNG and return the file-like object
    with PNG data.
    """
    def __init__(self, palette=False, colors=256, sharpness=None,
                 background='white'):
        self.palette = palette
        self.colors = colors
        self.background = background
        self.sharpness = sharpness

    def filter(self, im):
        if self.sharpness:
            im = sharpen(im, self.sharpness)
        buf = BytesIO()
        if self.palette:
            if im.mode in ('RGBA', 'LA'):
                alpha = im.split()[3]
                alpha = Image.eval(alpha, lambda a: 255 if a > 0 else 0)
                mask = Image.eval(alpha, lambda a: 255 if a == 0 else 0)

                matte = Image.new("RGBA", im.size, self.background)
                matte.paste(im, (0, 0), im)
                matte = matte.convert("RGB").convert(
                    'P', palette=Image.ADAPTIVE, colors=self.colors - 1)
                matte.paste(self.colors, mask)
                matte.save(buf, "PNG", transparency=self.colors)
            elif im.mode not in ('P'):
                im = im.convert('P', palette=Image.ADAPTIVE,
                                colors=self.colors)
                im.save(buf, 'PNG')
            else:
                im.save(buf, 'PNG')
        else:
            if not im.mode.startswith("RGB"):
                im = im.convert('RGB')
            im.save(buf, 'PNG')
        buf.seek(0)
        return buf


class PNGProcessor(Filter):
    """
    Postprocess a PNG. For now, just uses pngcrush and optipng to do some
    additional lossless slimming.
    """

    def __call__(self, input):
        input = self.shell_process(input, ['pngcrush', 'IN', 'OUT'])
        return self.shell_process(input, ['optipng', 'IN'], inplace=True)



from __future__ import absolute_import, print_function, division

import re
import posixpath
import os.path

from mako.exceptions import TopLevelLookupException
from mako.lookup import TemplateLookup


class SuperTemplateLookup(TemplateLookup):
    """
    A subclass of ``mako.lookup.TemplateLookup`` which provides an extra bit of
    functionality for building template inheritance chains.

    For normal documented usage, behavior is the same as the normal
    ``TemplateLookup`` class. However, it is additional possible for a template
    to inherit from another template which has the same file path, by prefixing
    the URI with ``super:``. This inheritance style is not possible with the
    base ``TemplateLookup``.
    """
    super_delimiter = '::'

    def find_dir_for_path(self, filename):
        """
        Given a file path, return the template directory within the lookup's
        list of directories where it can be found.
        """
        for dir in self.directories:
            if os.path.commonprefix([filename, dir]) == dir:
                return dir
        raise ValueError("filename %r is not in any of this lookup's "
                         "directories" % filename)

    def find_dir_for_uri(self, uri, start_index=0):
        """
        Given a template URI, return the first directory in the lookup's list
        of directories where it can be found, starting at the ``start_index``
        directory in the list.
        """
        u = re.sub(r'^\/+', '', uri)
        for dir in self.directories[start_index:]:
            srcfile = posixpath.normpath(posixpath.join(dir, u))
            if os.path.isfile(srcfile):
                return dir, srcfile
        return None, None

    def get_template(self, uri):
        """
        Fetch a template from this lookup.

        If uri containers the ``super_delimiter``, expect it to join a 2-tuple
        of ``(uri, dir)``. The ``dir`` component refers to the specific
        directory within the lookup that the template should be fetched from.
        """
        try:
            if self.filesystem_checks:
                return self._check(uri, self._collection[uri])
            else:
                return self._collection[uri]

        except KeyError:

            if self.super_delimiter in uri:
                file_path, from_dir = uri.split(self.super_delimiter)
                u = re.sub(r'^\/+', '', file_path)
                srcfile = posixpath.normpath(posixpath.join(from_dir, u))
                return self._load(srcfile, uri)

            else:
                dir, srcfile = self.find_dir_for_uri(uri)
                if not dir:
                    raise TopLevelLookupException(
                        "Cant locate template for uri %r" % uri)
                return self._load(srcfile, uri)

    def adjust_uri(self, uri, relativeto):
        """
        Adjust the supplied URI given a specific source directory. If the uri
        begins with ``super:``, relativeize the URI such that it refers to the
        URI specified in the next-higher template directory in the chain.
        """
        if uri.startswith('super:'):
            uri = uri[6:]
            relative_templ = self.get_template(relativeto)
            from_dir = self.find_dir_for_path(relative_templ.filename)
            assert from_dir, "can't find what dir calling template is in"
            start_index = self.directories.index(from_dir) + 1
            relpath = os.path.relpath(relative_templ.filename, from_dir)

            next_dir, srcfile = self.find_dir_for_uri(relpath, start_index)
            assert next_dir
        else:
            next_dir = None

        if self.super_delimiter in relativeto:
            relativeto = relativeto.split(self.super_delimiter)[0]

        uri = TemplateLookup.adjust_uri(self, uri, relativeto)
        if next_dir:
            assert self.super_delimiter not in uri
            new_uri = uri + self.super_delimiter + next_dir
            return new_uri
        else:
            return uri

    def filename_to_uri(self, filename):
        """
        Given a filename, return a URI which can be used to access this
        template from the lookup.
        """
        from_dir = self.find_dir_for_path(filename)
        base_uri = TemplateLookup.filename_to_uri(self, filename)
        ret = base_uri + self.super_delimiter + from_dir
        return ret



from __future__ import absolute_import, print_function, division

import os.path
import inspect
import pkg_resources

from pyramid.decorator import reify
from pyramid.settings import aslist, asbool
from pyramid.path import DottedNameResolver

from .templating.lookup import SuperTemplateLookup
from .templating.renderer import (mako_renderer_factory,
                                  mako_renderer_factory_txt)

static_dir = pkg_resources.resource_filename('pyramid_frontend', 'static')


default_sentinel = object()


class Theme(object):
    """
    Represents a collection of templates, static files, image filters, and
    configuration corresponding to a particular visual theme (or "skin") used
    by the application.

    New themes are created by subclassing from this class. When passed to
    ``config.add_theme()``, The subclass will be instantiated with the
    application's ``settings`` dict and prepared for use.
    """
    template_dir = 'templates'
    static_dir = 'static'
    assets = {}
    image_filters = []
    includes = []

    cache_impl = None
    cache_args = None

    def __init__(self, settings):
        self.settings = settings
        self._compiled_asset_cache = {}

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.key)

    @classmethod
    def qualify_path(cls, path):
        theme_file = os.path.abspath(inspect.getfile(cls))
        return os.path.join(os.path.dirname(theme_file), path)

    @classmethod
    def traverse_attributes(cls, name, qualify_paths=False):
        assert len(cls.__bases__) == 1, \
            "multiple inheritance not allowed for themes"
        while cls != Theme:
            el = getattr(cls, name)
            if qualify_paths:
                el = cls.qualify_path(el)
            yield cls.key, el
            cls = cls.__bases__[0]

    @reify
    def template_dirs(self):
        dirs = []
        for key, dir in self.__class__.traverse_attributes(
                'template_dir', qualify_paths=True):
            dirs.append(dir)
        return dirs

    @reify
    def lookup(self):
        return self._make_lookup()

    @reify
    def lookup_nofilters(self):
        return self._make_lookup(clear_default_filters=True)

    def _make_lookup(self, clear_default_filters=False):
        default_filters = (['decode.utf8']
                           if clear_default_filters else
                           ['escape'])

        template_imports = [
            'from webhelpers2.html import escape',
        ]
        template_imports.extend(aslist(
            self.settings.get('pyramid_frontend.template_imports', ''),
            flatten=False))

        debug = asbool(self.settings.get('pyramid_frontend.debug'))
        base_module_dir = \
            self.settings.get('pyramid_frontend.module_directory')
        module_dir = base_module_dir and os.path.join(base_module_dir,
                                                      self.key)

        return SuperTemplateLookup(directories=self.template_dirs,
                                   input_encoding='utf-8',
                                   output_encoding='utf-8',
                                   imports=template_imports,
                                   default_filters=default_filters,
                                   filesystem_checks=debug,
                                   module_directory=module_dir,
                                   cache_impl=self.cache_impl,
                                   cache_args=self.cache_args)

    @reify
    def stacked_image_filters(self):
        filters = {}
        collected = self.__class__.traverse_attributes('image_filters')
        for key, class_list in reversed(list(collected)):
            class_dict = {chain.suffix: chain for chain in class_list}
            filters.update(class_dict)
        return filters.values()

    @reify
    def stacked_assets(self):
        asset_specs = {}
        collected = self.__class__.traverse_attributes('assets')
        for key, class_dict in reversed(list(collected)):
            asset_specs.update(class_dict)
        return asset_specs

    @reify
    def stacked_includes(self):
        includes = []
        collected = self.__class__.traverse_attributes('includes')
        for key, these in reversed(list(collected)):
            includes.extend(these)
        return includes

    @reify
    def keyed_static_dirs(self):
        cls = self.__class__
        stack = list(cls.traverse_attributes('static_dir', qualify_paths=True))
        stack.append(('pfe', static_dir))
        return stack

    def compiled_asset_path(self, key):
        if key in self._compiled_asset_cache:
            return self._compiled_asset_cache[key]
        else:
            map_path = os.path.join(
                self.settings['pyramid_frontend.compiled_asset_dir'],
                self.key,
                '%s.map' % key)
            with open(map_path) as f:
                self._compiled_asset_cache[key] = compiled_path = f.read()
                return compiled_path

    def static_url_to_filesystem_path(self, url):
        """
        Given a URL of the structure /_<theme key>/<path>, locate the static
        dir which corresponds to the theme key and re-qualify the <path> to
        that directory.
        """
        assert url.startswith('/_')
        theme_key, path = url[2:].split('/', 1)
        theme_dirs = dict(self.keyed_static_dirs)
        base_dir = theme_dirs[theme_key]
        return os.path.join(base_dir, path)

    def opt(self, key, default=default_sentinel):
        if default is default_sentinel:
            return getattr(self, key)
        else:
            return getattr(self, key, default)

    def static(self, path):
        for key, static_dir in self.keyed_static_dirs:
            if os.path.exists(os.path.join(static_dir, path)):
                return '/_%s/%s' % (key, path)
        raise IOError('path %r does not exist in any static dirs' % path)

    def compile(self, minify=True):
        output_dir = os.path.join(
            self.settings['pyramid_frontend.compiled_asset_dir'],
            self.key)
        for key, asset in self.stacked_assets.items():
            asset.compile(key=key,
                          theme=self,
                          output_dir=output_dir,
                          minify=minify)


def add_theme(config, cls):
    """
    A Pyramid config directive to initialiaze and register a theme for use.
    """
    resolved_cls = config.maybe_dotted(cls)

    settings = config.registry.settings
    theme = resolved_cls(settings)

    # Call includes
    package = inspect.getmodule(resolved_cls)
    resolver = DottedNameResolver(package=package)
    for include in theme.stacked_includes:
        config.include(resolver.maybe_resolve(include))

    # Register static dirs.
    static_dirs = settings.setdefault('pyramid_frontend.static_registry',
                                      set())
    for key, dir in theme.keyed_static_dirs:
        if (key, dir) not in static_dirs:
            static_dirs.add((key, dir))
            config.add_static_view('_%s' % key, path=dir)

    # Update global image filter registry as well, and ensure there are no
    # conflicts.
    for chain in theme.stacked_image_filters:
        config.add_image_filter(chain, with_theme=theme)

    def register(theme):
        themes = settings.setdefault('pyramid_frontend.theme_registry', {})
        themes[theme.key] = theme

    intr = config.introspectable(category_name='themes',
                                 discriminator=theme.key,
                                 title=theme.key,
                                 type_name=None)
    intr['theme'] = theme

    config.action(('theme', theme.key),
                  register,
                  args=(theme,),
                  introspectables=(intr,))


def set_theme_strategy(config, strategy_func):
    """
    A Pyramid config directive to set a customized theme-selection strategy for
    each request.
    """
    def register():
        registry = config.registry
        registry.pfe_theme_strategy = strategy_func

    config.action(('theme_strategy',), register)


def default_theme_strategy(request):
    """
    The default theme selection strategy: just checks the
    ``pyramid_frontend.theme`` settings key.
    """
    settings = request.registry.settings
    return settings['pyramid_frontend.theme']


def theme(request):
    """
    The theme instance that should be used for this request. This property is
    both lazily-evaluated and reified.
    """
    registry = request.registry
    strategy = getattr(registry, 'pfe_theme_strategy', default_theme_strategy)
    key = strategy(request)
    settings = registry.settings
    themes = settings.setdefault('pyramid_frontend.theme_registry', {})
    return themes[key]


def includeme(config):
    config.include('.images')
    config.include('.assets')

    config.add_directive('add_theme', add_theme)
    config.add_directive('set_theme_strategy', set_theme_strategy)

    config.add_request_method(theme, 'theme', reify=True)

    config.add_renderer(name='.html', factory=mako_renderer_factory)
    config.add_renderer(name='.txt', factory=mako_renderer_factory_txt)



from __future__ import absolute_import, print_function, division

from ..base import BaseTheme
from pyramid_frontend.images.chain import FilterChain


class BarTheme(BaseTheme):
    key = 'bar'
    image_filters = [
        FilterChain('barthumb', width=300, height=100, crop=True),
    ]
    includes = [
        '..base.base_includeme',
    ]



from __future__ import absolute_import, print_function, division

import os
import stat

from shutil import copyfileobj

from .files import filter_sep
from .filters import (PNGSaver, PNGProcessor, JPGSaver, JPGProcessor,
                      ThumbFilter)

savers = {
    'png': PNGSaver,
    'jpg': JPGSaver,
}


postprocessors = {
    'png': PNGProcessor,
    'jpg': JPGProcessor,
}


class FilterChain(object):
    """
    A chain of image filters (a.k.a. "pipeline") used to process images for a
    particular display context.
    """
    def __init__(self, suffix, filters=(), extension='png',
                 width=None, height=None, no_thumb=False,
                 pad=False, crop=False, crop_whitespace=False,
                 background='white', enlarge=False,
                 **saver_kwargs):

        self.suffix = suffix
        self.filters = list(filters)
        self.width = width
        self.height = height
        self.extension = extension

        assert filter_sep not in suffix, \
            "filter suffix cannot contain %r" % filter_sep
        assert suffix.lower() == suffix, \
            "filter suffix must be lowercase"

        if not no_thumb:
            self.filters.append(ThumbFilter(
                (width, height),
                pad=pad, crop=crop, crop_whitespace=crop_whitespace,
                background=background, enlarge=enlarge))

        saver_class = savers[self.extension]
        self.filters.append(saver_class(**saver_kwargs))

        self.filters.append(postprocessors[self.extension]())

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.suffix)

    def basename(self, name, original_ext):
        return ''.join([name,
                        filter_sep,
                        original_ext,
                        filter_sep,
                        self.suffix,
                        '.',
                        self.extension])

    def run_chain(self, image_data):
        for filter in self.filters:
            image_data = filter(image_data)
        return image_data

    def write(self, dest_path, filtered):
        # We have the final image data, now save it.
        dest_dir = os.path.dirname(dest_path)
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)
        copyfileobj(filtered, open(dest_path, 'wb'))

        # Make this writable by everyone.
        bits = os.stat(dest_path).st_mode
        os.chmod(dest_path, bits | stat.S_IWGRP | stat.S_IWOTH)

        return True

    def run(self, dest_path, image_data):
        filtered = self.run_chain(image_data)
        return self.write(dest_path, filtered)


class PassThroughFilterChain(FilterChain):
    """
    A filter chain which does not do any manipulation, only applies lossless
    optimization tools.
    """
    def __init__(self, suffix=None, filters=()):
        self.suffix = suffix
        self.extension = None
        self.filters = filters
        self.width = None
        self.height = None

    def basename(self, name, original_ext):
        return '%s.%s' % (name, original_ext)

    def run(self, dest_path, image_data):
        filtered = self.run_chain(image_data)
        ext = dest_path.rsplit('.', 1)[-1]
        if ext in postprocessors:
            filtered = postprocessors[ext]()(filtered)
        return self.write(dest_path, filtered)



from __future__ import absolute_import, print_function, division

import sys

from mako import exceptions
from pyramid.compat import reraise


class MakoRenderingException(Exception):
    """
    When an exception occurs during the rendering of a template, it will be
    re-raised as one of these exceptions.
    """
    def __init__(self, s):
        self.s = s

    def __repr__(self):
        return self.s

    __str__ = __repr__


class MakoRenderer(object):
    """
    A class which can be instantiated to make a Pyramid renderer for Mako
    templates.
    """

    def __init__(self, info, plaintext=False):
        self.name = info.name
        self.plaintext = plaintext

    def __call__(self, value, system):
        """Find and render a template.

        This replaces the default Mako renderer to locate templates by theme.

        ``system`` will contain default renderer globals plus those added by
        by :func:`add_renderer_globals`.

        ``value`` will be either the dict returned from a view or the dict
        passed to one of Pyramid's render functions. These values will
        override the ``system`` values.

        :func:`add_renderer_globals` will be called to update ``system``
        before this method is called; it can access ``value`` via
        ``event.rendering_val``.

        """
        # Avoid conflict between Pyramid and Mako ``context``s
        system['_context'] = system.pop('context', None)

        try:
            system.update(value)
        except (TypeError, ValueError):
            raise ValueError('renderer was passed non-dictionary as value')

        request = system['request']

        theme = request.theme
        if self.plaintext:
            lookup = theme.lookup_nofilters
            response = request.response
            ct = response.content_type
            if ct == response.default_content_type:
                response.content_type = 'text/plain'
        else:
            lookup = theme.lookup

        template = lookup.get_template(self.name)

        try:
            result = template.render_unicode(**system)
        except:
            try:
                exc_info = sys.exc_info()
                errtext = exceptions.text_error_template().render(
                    error=exc_info[1],
                    traceback=exc_info[2])
                reraise(MakoRenderingException(errtext), None, exc_info[2])
            finally:
                del exc_info

        return result


def mako_renderer_factory(info):
    """
    A Pyramid renderer factory to render Mako templates with default settings.
    """
    return MakoRenderer(info)


def mako_renderer_factory_txt(info):
    """
    A Pyramid renderer factory to render Mako templates with no default
    HTML-escaping filters. This renderer factory should generally be used for
    plaintext templates rather than HTML.
    """
    return MakoRenderer(info, plaintext=True)



from __future__ import absolute_import, print_function, division

import math

from PIL import Image, ImageChops, ImageEnhance


def flatten_alpha(im, background='white'):
    "Composite an RGBA image with a flat background."
    if im.mode == "RGBA":
        ni = Image.new("RGB", im.size, background)
        ni.paste(im, im)
        im = ni
    return im


def pad_image(im, dimensions, mode=None, color=None):
    """
    Pad an image to a given set of dimensions.
    """
    w, h = im.size
    desired_w, desired_h = (max(a, b) for a, b in zip(im.size, dimensions))
    new = Image.new(mode or 'RGB', (desired_w, desired_h), color or 'white')
    woff = int((desired_w - w) / 2)
    hoff = int((desired_h - h) / 2)
    new.paste(im, (woff, hoff))
    return new


def image_entropy(im):
    hist = im.histogram()
    hist_size = sum(hist)
    # normalize
    hist = [float(h) / hist_size for h in hist]
    return -sum([p * math.log(p, 2) for p in hist if p != 0])


def crop_entropy_width(im, desired_w):
    w, h = im.size
    while w > desired_w:
        # Remove slice from left or right side, whichever has less entropy.
        slice_width = min(w - desired_w, int(math.floor(w * 0.1)))
        left = im.crop((0, 0, slice_width, h))
        right = im.crop((w - slice_width, 0, w, h))

        if image_entropy(left) < image_entropy(right):
            # Crop off left side.
            im = im.crop((slice_width, 0, w, h))
        else:
            # Crop off right side.
            im = im.crop((0, 0, w - slice_width, h))

        w, h = im.size

    return im


def crop_entropy_height(im, desired_h):
    w, h = im.size
    while h > desired_h:
        # Remove slice from top or bottom, whichever has less entropy.
        slice_height = min(h - desired_h, int(math.floor(h * 0.1)))
        top = im.crop((0, 0, w, slice_height))
        bottom = im.crop((0, h - slice_height, w, h))

        if image_entropy(bottom) < image_entropy(top):
            im = im.crop((0, 0, w, h - slice_height))
        else:
            im = im.crop((0, slice_height, w, h))

        w, h = im.size

    return im


def crop_entropy(im, dimensions):
    """
    Scale and crop the image to the desired aspect ratio by slicing off a bit
    at a time.  Slice off the edge which is lowest entropy.
    """
    # First resize this so that the longer dimension matches the desired
    # dimension.
    input_w, input_h = im.size
    desired_w, desired_h = dimensions

    input_ar = float(input_w) / float(input_h)
    desired_ar = float(desired_w) / float(desired_h)

    if input_ar > desired_ar:
        scale_h = desired_h
        scale_w = int(input_h * input_ar) + 1
        im.thumbnail((scale_w, scale_h), Image.ANTIALIAS)
        return crop_entropy_width(im, desired_w)
    elif input_ar < desired_ar:
        scale_w = desired_w
        scale_h = int(input_w / input_ar) + 1
        im.thumbnail((scale_w, scale_h), Image.ANTIALIAS)
        return crop_entropy_height(im, desired_h)
    else:
        # AR already matches exactly, no need to crop.
        im.thumbnail((desired_w, desired_h), Image.ANTIALIAS)
        return im


def colors_differ(a, b, tolerance):
    manhattan_distance = sum([abs(i - j) for i, j in zip(a, b)])
    return manhattan_distance > tolerance


def is_white_background(im, tolerance=180):
    """
    Check if this image is against a white background: that is, if every border
    pixel is white.
    """
    w, h = im.size
    im = im.convert('RGB')
    array = im.load()
    white = (255, 255, 255)
    for x in range(w):
        # Top border
        if colors_differ(array[x, 0], white, tolerance):
            return False
        # Bottom border
        if colors_differ(array[x, h - 1], white, tolerance):
            return False
    for y in range(h):
        # Left border
        if colors_differ(array[0, y], white, tolerance):
            return False
        # Right border
        if colors_differ(array[w - 1, y], white, tolerance):
            return False
    return True


def is_larger(im, dimensions):
    """
    Check if this image is currently larger in either axis than the desired
    dimensions. If either desired dimension is None, the image will be larger
    in that dimension.
    """
    w, h = im.size
    desired_w, desired_h = dimensions
    return (w > (desired_w or 0)) or (h > (desired_h or 0))


def bounding_box(im, tolerance=0):
    """
    A bounding box algorithm that has some tolerance for fuzziness.
    """
    bg = Image.new(im.mode, im.size, im.getpixel((0, 0)))
    diff = ImageChops.difference(im, bg)
    diff = ImageChops.add(diff, diff, 2.0, -tolerance)
    return diff.getbbox()


def sharpen(im, sharpness):
    sharpener = ImageEnhance.Sharpness(im)
    return sharpener.enhance(sharpness)



from __future__ import absolute_import, print_function, division

import os
import re
import io

from webhelpers2.html.tags import HTML

import six

from .. import cmd
from .asset import Asset


class LessAsset(Asset):
    """
    Asset handler for LESS CSS files. In production, autoprefixer is used.
    """
    extension = 'css'

    import_re = re.compile(r'@import[ ]+"(?P<path>.*)";')

    def __init__(self, url_path,
                 less_path='/_pfe/less.js',
                 lessc_path='lessc',
                 postcss_path='postcss'):
        self.url_path = url_path
        self.less_path = less_path
        self.lessc_path = lessc_path
        self.postcss_path = postcss_path

    def compile(self, key, theme, output_dir, minify=True):
        """
        Compile a LESS entry point.
        """
        entry_point = theme.static_url_to_filesystem_path(self.url_path)

        lessc_flags = [self.lessc_path]
        lessc_flags.append('--verbose')
        if minify:
            lessc_flags.append('--compress')

        preprocessed = self.concatenate(theme, entry_point)
        assert isinstance(preprocessed, six.text_type)

        preprocessed = preprocessed.encode('utf-8')

        with self.tempfile() as (inter_f, inter_name):
            with self.tempfile() as (in_f, in_name):
                in_f.write(preprocessed)
                in_f.flush()
                lessc_cmd = lessc_flags + [in_name]
                lessc_cmd.append(inter_name)
                cmd.run(lessc_cmd)

            with self.tempfile() as (out_f, out_name):
                postcss_cmd = [self.postcss_path,
                               '--use', 'autoprefixer',
                               '--output', out_name,
                               inter_name]
                cmd.run(postcss_cmd)

                file_path = self.write_from_file(key, out_name, entry_point,
                                                 output_dir)

        return file_path

    def concatenate(self, theme, start_path):
        """
        Combine a LESS file and its `@import`s, recursively. Used to keep
        the ``lessc`` command-line compiler from having to traverse a directory
        structure, which it doesn't support very well right now.

        Always assumes file encodings are UTF-8, and should always return a
        unicode string (unicode on 2.x, str on 3.x).
        """
        err = 'File does not exist {0}'.format(start_path)
        assert os.path.isfile(start_path), err
        contents = []
        directory = os.path.dirname(start_path)
        with io.open(start_path, encoding='utf8') as fp:
            for line in fp:
                match = self.import_re.match(line.strip())
                if match:
                    path = match.groupdict()['path']
                    ext = os.path.splitext(path)[1]
                    if ext not in ('.css', '.less'):
                        path = '.'.join((path, 'less'))
                    if os.path.isabs(path):
                        path = theme.static_url_to_filesystem_path(path)
                    else:
                        path = os.path.join(directory, path)
                    contents.append(self.concatenate(theme, path))
                else:
                    contents.append(line)
        return u''.join(contents)

    def tag_development(self, theme, url):
        """
        Return an HTML fragment to use a less CSS entry point in development.
        """
        return ''.join([
            HTML.link(rel='stylesheet/less', type='text/css', href=url),
            HTML.script(src=self.less_path),
        ])

    def tag_production(self, theme, url):
        """
        Return an HTML fragment to use a less CSS entry point in production.
        """
        return HTML.link(rel='stylesheet', type='text/css', href=url)



from __future__ import absolute_import, print_function, division

import os
import os.path
import pkg_resources

from unittest import TestCase

from PIL import Image

from ..images import filters

samples_dir = pkg_resources.resource_filename('pyramid_frontend.tests', 'data')


def filesize(f):
    orig_pos = f.tell()
    f.seek(0, os.SEEK_END)
    size = f.tell()
    f.seek(orig_pos)
    return size


class TestFilters(TestCase):
    def assertSimilarColor(self, a, b, threshold=20):
        # Check manhattan distance
        if hasattr(a, '__len__'):
            dist = sum(abs(x - y) for x, y in zip(a, b))
        else:
            dist = abs(a - b)
        self.assertLess(dist, threshold)

    def test_png_save_rgb(self):
        saver = filters.PNGSaver(palette=True)
        im = Image.new("RGB", (25, 25), "red")
        f = saver(im)

        nm = Image.open(f)
        self.assertEqual(nm.mode, "P")
        nm = nm.convert("RGBA")
        self.assertEqual(nm.getpixel((8, 8)), (255, 0, 0, 255))

    def test_png_save_palette(self):
        saver = filters.PNGSaver(palette=True)
        im = Image.new("P", (25, 25))
        im.putpalette([0] * 768)
        im.paste(23, (0, 0, 25, 25))
        f = saver(im)

        nm = Image.open(f)
        self.assertEqual(nm.mode, "P")
        self.assertEqual(nm.getpixel((8, 8)), 23)

    def test_png_save_alpha_palette(self):
        saver = filters.PNGSaver(palette=True)
        im = Image.new("RGBA", (25, 25), (255, 0, 0))
        im.putpixel((10, 10), (0, 255, 0))

        f = saver(im)
        nm = Image.open(f)
        self.assertEqual(nm.mode, "P")
        converted = nm.convert('RGB')
        self.assertEqual(converted.getpixel((10, 10)), (0, 255, 0))

    def test_png_save_monochrome(self):
        saver = filters.PNGSaver()
        im = Image.new('L', (25, 25), 127)
        im.putpixel((10, 10), 255)

        f = saver(im)
        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGB')
        self.assertEqual(nm.getpixel((10, 10)), (255, 255, 255))

    def test_png_postprocess(self):
        saver = filters.PNGSaver()
        processor = filters.PNGProcessor()
        im = Image.new("RGBA", (25, 25), (255, 0, 0))
        im.putpixel((10, 10), (0, 255, 0))

        f = saver(im)
        nm = Image.open(f)
        self.assertEqual(nm.mode, "RGBA")
        self.assertEqual(nm.getpixel((10, 10)), (0, 255, 0, 255))

        size_before = filesize(f)

        f = processor(f)
        size_after = filesize(f)

        self.assertGreater(size_after, 0)
        self.assertLessEqual(size_after, size_before)

        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGBA')
        self.assertEqual(nm.size, (25, 25))

    def test_png_save_sharpen(self):
        saver = filters.PNGSaver(sharpness=1.5)
        im = Image.new('RGBA', (25, 25), 'white')
        im.putpixel((10, 10), (255, 127, 127, 127))

        f = saver(im)
        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGBA')
        self.assertEqual(nm.getpixel((10, 10)), (255, 88, 88, 127))

    def test_file_save(self):
        saver = filters.PNGSaver()
        f = open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'), 'rb')
        f = saver(f)

        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGB')
        self.assertEqual(nm.size, (512, 512))

    def test_jpeg_save(self):
        saver = filters.JPGSaver()
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        f = saver(im)

        nm = Image.open(f)
        self.assertEqual(nm.mode, "RGB")
        self.assertEqual(nm.getpixel((300, 300)), (206, 205, 1))

    def test_jpeg_save_low_quality(self):
        saver = filters.JPGSaver(quality=20)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        f = saver(im)

        # Should be less than 20K or so.
        self.assertLess(filesize(f), 20000)

    def test_jpeg_cmyk_save_png(self):
        saver = filters.PNGSaver()
        f = open(os.path.join(samples_dir, 'smiley-jpeg-cmyk.jpg'), 'rb')
        f = saver(f)

        nm = Image.open(f)
        self.assertEqual(nm.mode, "RGB")
        self.assertEqual(nm.size, (512, 512))

    def test_jpeg_cmyk_load_directly(self):
        # TODO May want to change this behavior at some point, and handle this
        # case.
        saver = filters.PNGSaver()
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-cmyk.jpg'))

        with self.assertRaises(AssertionError):
            saver(im)

    def test_jpeg_save_alpha(self):
        saver = filters.JPGSaver(quality=95)
        im = Image.new('RGBA', (30, 30), (255, 0, 0, 0))
        rect = Image.new('RGBA', (10, 10), (0, 255, 0, 255))
        im.paste(rect, (10, 10))

        f = saver(im)
        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGB')
        self.assertSimilarColor(nm.getpixel((15, 15)), (0, 255, 0))

    def test_jpeg_save_monochrome(self):
        saver = filters.JPGSaver(quality=95)
        im = Image.new('L', (30, 30), 255)
        rect = Image.new('L', (10, 10), 0)
        im.paste(rect, (10, 10))

        f = saver(im)
        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGB')
        self.assertEqual(nm.getpixel((15, 15)), (0, 0, 0))

    def test_jpeg_postprocess(self):
        saver = filters.JPGSaver()
        processor = filters.JPGProcessor()
        im = Image.new("RGB", (25, 25), (255, 0, 0))
        im.putpixel((10, 10), (0, 255, 0))

        f = saver(im)
        size_before = filesize(f)

        f = processor(f)
        size_after = filesize(f)

        self.assertGreater(size_after, 0)
        self.assertLessEqual(size_after, size_before)

        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGB')
        self.assertEqual(nm.size, (25, 25))

    def test_jpeg_save_sharpen(self):
        saver = filters.JPGSaver(sharpness=1.5)
        im = Image.new('RGB', (30, 30), 'white')
        rect = Image.new('RGB', (10, 10), (255, 0, 0))
        im.paste(rect, (10, 10))

        f = saver(im)
        nm = Image.open(f)
        self.assertEqual(nm.mode, 'RGB')
        self.assertSimilarColor(nm.getpixel((15, 15)), (254, 6, 0))

    def test_vignette_filter(self):
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        filter = filters.VignetteFilter()
        saver = filters.JPGSaver()

        im = filter(im)
        f = saver(im)

        nm = Image.open(f)
        self.assertNotEqual(nm.getpixel((0, 0)), (255, 255, 255))

    def test_thumb_filter_defaults(self):
        filter = filters.ThumbFilter((64, 32))
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (32, 32))

    def test_thumb_filter_crop(self):
        filter = filters.ThumbFilter((64, 32), crop=True)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (64, 32))

    def test_thumb_filter_pad(self):
        filter = filters.ThumbFilter((64, 32), pad=True)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (64, 32))

    def test_thumb_filter_crop_nonwhite(self):
        filter = filters.ThumbFilter((64, 32), crop='nonwhite')
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        # This test image has a white background, so it should just get
        # resized, and not cropped.
        self.assertEqual(im.size, (32, 32))

    def test_thumb_filter_crop_nonwhite_background(self):
        filter = filters.ThumbFilter((64, 32), crop='nonwhite')
        im = Image.new('RGB', (100, 100), 'red')
        im = filter(im)
        # This test image has a nonwhite background, so it should get cropped.
        self.assertEqual(im.size, (64, 32))

    def test_thumb_filter_crop_whitespace(self):
        filter = filters.ThumbFilter((64, 32), crop_whitespace=True)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (32, 32))

    def test_thumb_filter_enlarge(self):
        filter = filters.ThumbFilter((1000, 1000), enlarge=True)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (1000, 1000))

    def test_thumb_filter_unspecified_dims(self):
        filter = filters.ThumbFilter((None, None), crop_whitespace=True)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (488, 488))

    def test_thumb_filter_unspecified_width(self):
        filter = filters.ThumbFilter((None, 100), crop_whitespace=True)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (100, 100))

    def test_thumb_filter_unspecified_height(self):
        filter = filters.ThumbFilter((200, None), crop_whitespace=True)
        im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))
        im = filter(im)
        self.assertEqual(im.size, (200, 200))



from __future__ import absolute_import, print_function, division

from webhelpers2.html.tags import HTML

from .files import (prefix_for_name, get_url_prefix, original_path,
                    save_image, save_to_error_dir, check, filter_sep)
from .view import ImageView, MissingOriginal
from .chain import PassThroughFilterChain, FilterChain

__all__ = ['FilterChain', 'MissingOriginal',
           'save_image', 'save_to_error_dir', 'check', 'filter_sep']


def add_image_filter(config, chain, with_theme=None):
    """
    Pyramid config directive to register an image filter chain.
    """
    def register(with_theme):
        registry = config.registry

        if not hasattr(registry, 'image_filter_registry'):
            registry.image_filter_registry = {}
        filter_registry = registry.image_filter_registry

        if chain.suffix in filter_registry:
            registered_chain, registered_theme_set = \
                filter_registry[chain.suffix]
            if registered_chain != chain:
                raise ValueError(
                    "suffix %r already registered with different instance" %
                    chain.suffix)
            registered_theme_set.add(with_theme)
        else:
            with_theme_set = set()
            if with_theme:
                with_theme_set.add(with_theme)
            filter_registry[chain.suffix] = (chain, with_theme_set)

    intr = config.introspectable(category_name='image_filters',
                                 discriminator=chain.suffix,
                                 title=chain.suffix,
                                 type_name=None)
    intr['chain'] = chain
    intr['with_theme'] = with_theme

    config.action(('image_filter', chain.suffix, with_theme),
                  register,
                  args=(with_theme,),
                  introspectables=(intr,))


def image_url(request, name, original_ext, filter_key,
              qualified=False, _scheme=None, _host=None, _port=None):
    """
    Return the URL for an image as processed by a specified image filter chain.
    """
    filter_registry = request.registry.image_filter_registry

    # Check if there is a theme active. If so, check that the supplied
    # filter_key is ref'd within the theme: if not, fail with a descriptive
    # exception.
    chain, with_theme_set = filter_registry[filter_key]
    if with_theme_set and getattr(request, 'theme', None):
        assert request.theme in with_theme_set, \
            ("current theme is %r, but this filter is only registered "
             "with %r" % (request.theme, with_theme_set))

    prefix = prefix_for_name(name)
    name = chain.basename(name, original_ext)
    if qualified:
        return request.route_url('pyramid_frontend:images',
                                 prefix=prefix,
                                 name=name,
                                 _scheme=_scheme or request.scheme,
                                 _host=_host or request.host,
                                 _port=_port or request.server_port)
    else:
        return request.route_path('pyramid_frontend:images',
                                  prefix=prefix,
                                  name=name)


def image_tag(request, name, original_ext, filter_key,
              qualified=False, _scheme=None, _host=None, _port=None, **kwargs):
    """
    Return the HTML tag for an image as processed by a specified image filter
    chain.
    """
    filter_registry = request.registry.image_filter_registry
    chain, with_theme = filter_registry[filter_key]

    kwargs.setdefault('width', chain.width)
    kwargs.setdefault('height', chain.height)

    url = request.image_url(name, original_ext, filter_key,
                            qualified=qualified, _scheme=_scheme, _host=_host)

    return HTML.img(src=url, **kwargs)


def image_original_path(request, name, original_ext):
    """
    Return the filesystem path for an original image.
    """
    settings = request.registry.settings
    return original_path(settings, name, original_ext)


def includeme(config):
    config.add_directive('add_image_filter', add_image_filter)
    config.add_image_filter(PassThroughFilterChain())

    config.add_request_method(image_url, 'image_url')
    config.add_request_method(image_tag, 'image_tag')
    config.add_request_method(image_original_path, 'image_original_path')

    url_prefix = get_url_prefix(config.registry.settings)
    config.add_route('pyramid_frontend:images',
                     '%s/{prefix}/{name:.+\.\w+}' % url_prefix)
    config.add_view(ImageView, route_name='pyramid_frontend:images')



from __future__ import absolute_import, print_function, division

import os.path
import mimetypes
import pkg_resources

from pyramid.httpexceptions import HTTPNotFound
from pyramid.response import Response
from pyramid.static import FileResponse
from pyramid.settings import asbool
from lockfile import FileLock

from .files import filter_sep, prefix_for_name, processed_path, original_path


plausible_extensions = set([
    'jpg',
    'jpeg',
    'gif',
    'png',
    'tif',
    'tiff',
    'ppm',
    'bmp',
    'tga',
])


def get_image_filter(registry, filter_key):
    filter_registry = getattr(registry, 'image_filter_registry', {})
    chain, with_theme = filter_registry[filter_key]
    return chain


class MissingOriginal(Exception):

    def __init__(self, path, chain):
        self.path = path
        self.chain = chain
        Exception.__init__(self, 'Missing file %s' % path)


def process_image(settings, name, original_ext, chain, overwrite=False):
    proc_path = processed_path(settings, name, original_ext, chain)
    if overwrite or (not os.path.exists(proc_path)):
        dest_dir = os.path.dirname(proc_path)
        try:
            os.makedirs(dest_dir)
        except OSError:
            pass

        lock = FileLock(proc_path + '.lock')
        with lock:
            if overwrite or (not os.path.exists(proc_path)):
                orig_path = original_path(settings, name, original_ext)
                if not os.path.exists(orig_path):
                    raise MissingOriginal(path=orig_path, chain=chain)
                image_data = open(orig_path, 'rb')
                chain.run(proc_path, image_data)
    return proc_path


class ImageView(object):

    def __init__(self, request):
        self.request = request

    def placeholder(self, chain):
        orig_path = pkg_resources.resource_filename('pyramid_frontend.images',
                                                    'no-image.png')
        image_data = open(orig_path, 'rb')
        f = chain.run_chain(image_data)

        response = Response(f.read())
        response.content_type = \
            mimetypes.guess_type('x.%s' % chain.extension)[0]

        return response

    def __call__(self):
        request = self.request
        settings = request.registry.settings

        url_prefix = self.request.matchdict['prefix']
        name = self.request.matchdict['name']
        name, ext = name.rsplit('.', 1)

        if filter_sep in name:
            parts = name.split(filter_sep, 2)
            if len(parts) == 3:
                name, original_ext, chain_name = parts
            else:
                raise HTTPNotFound()
        else:
            original_ext = ext
            chain_name = None

        try:
            chain = get_image_filter(request.registry, chain_name)
        except KeyError:
            raise HTTPNotFound()

        if ((chain.extension and chain.extension != ext) or
                prefix_for_name(name) != url_prefix):
            raise HTTPNotFound()

        if original_ext not in plausible_extensions:
            raise HTTPNotFound()

        debug = asbool(settings.get('pyramid_frontend.debug'))
        overwrite = debug and request.params.get('overwrite')

        try:
            proc_path = process_image(settings, name, original_ext, chain,
                                      overwrite=overwrite)
        except MissingOriginal:
            if debug:
                return self.placeholder(chain)
            else:
                raise
        return FileResponse(proc_path, self.request)



from __future__ import absolute_import, print_function, division

from ..base import BaseTheme
from pyramid_frontend.images.chain import FilterChain
from pyramid_frontend.assets.less import LessAsset
from pyramid_frontend.assets.requirejs import RequireJSAsset
from pyramid_frontend.assets.svg import SVGAsset


def foo_includeme(config):
    test_includes = config.registry.settings.setdefault('test_includes', [])
    test_includes.append('foo')


class FooTheme(BaseTheme):
    key = 'foo'
    image_filters = [
        FilterChain('tiny', width=64, height=64, crop=True),
    ]
    assets = {
        'main-less': LessAsset('/_foo/css/main.less'),
        'main-js': RequireJSAsset('/_foo/js/main.js'),
        'logo-svg': SVGAsset('/_foo/images/logo.svg'),
    }
    includes = [
        '.foo_includeme',
    ]



from __future__ import absolute_import, print_function, division

import os
import os.path
import shutil
import pkg_resources

from unittest import TestCase, skip

from PIL import Image

from ..images.chain import FilterChain, PassThroughFilterChain

from . import utils

samples_dir = pkg_resources.resource_filename('pyramid_frontend.tests', 'data')


class TestFilterChain(TestCase):
    work_dir = os.path.join(utils.work_dir, 'filter-chain-tests')
    test_files = ['smiley-jpeg-rgb.jpg',
                  'smiley-png24-alpha.png',
                  'smiley-gif-alpha.gif',
                  'smiley-jpeg-cmyk.jpg']

    def setUp(self):
        if os.path.exists(self.work_dir):
            shutil.rmtree(self.work_dir)
        os.makedirs(self.work_dir)

    def test_basename(self):
        chain = FilterChain('thumb200', extension='png')
        self.assertEqual(chain.basename('some-cool-image', 'jpg'),
                         'some-cool-image_jpg_thumb200.png')
        self.assertEqual(chain.basename('delicious', 'gif'),
                         'delicious_gif_thumb200.png')

    def test_repr(self):
        chain = FilterChain('zygolicious', extension='png')
        self.assertIn('zygolicious', repr(chain))

    def test_basename_passthrough(self):
        chain = PassThroughFilterChain()
        self.assertEqual(chain.basename('some-cool-image', 'jpg'),
                         'some-cool-image.jpg')
        self.assertEqual(chain.basename('delicious', 'gif'),
                         'delicious.gif')

    def _process(self, chain, filename):
        orig_path = os.path.join(samples_dir, filename)
        image_data = open(orig_path, 'rb')
        proc_path = os.path.join(self.work_dir, filename)
        chain.run(proc_path, image_data)

        im = Image.open(proc_path)
        return im

    def _chain_basic(self, filename):
        chain = FilterChain('thumb50', extension='png',
                            width=50, height=50)
        im = self._process(chain, filename)
        self.assertEqual(im.size, (50, 50))

    def test_chain_basic_jpeg_rgb(self):
        self._chain_basic('smiley-jpeg-rgb.jpg')

    def test_chain_basic_jpeg_cmyk(self):
        self._chain_basic('smiley-jpeg-cmyk.jpg')

    def test_chain_basic_png(self):
        self._chain_basic('smiley-png24-alpha.png')

    # Pillow has broken GIF support in the versions that support py3k.
    @skip
    def test_chain_basic_gif(self):
        self._chain_basic('smiley-gif-alpha.gif')

    def _chain_passthrough(self, filename):
        chain = PassThroughFilterChain()
        im = self._process(chain, filename)
        self.assertEqual(im.size, (512, 512))

    def test_chain_passthrough_jpeg_rgb(self):
        self._chain_passthrough('smiley-jpeg-rgb.jpg')

    def test_chain_passthrough_jpeg_cmyk(self):
        self._chain_passthrough('smiley-jpeg-cmyk.jpg')

    def test_chain_passthrough_png(self):
        self._chain_passthrough('smiley-png24-alpha.png')

    def test_chain_passthrough_gif(self):
        self._chain_passthrough('smiley-gif-alpha.gif')

    def test_run_missing_output_dir(self):
        chain = FilterChain('thumb50', extension='png',
                            width=50, height=50)
        filename = self.test_files[0]
        orig_path = os.path.join(samples_dir, filename)
        image_data = open(orig_path, 'rb')
        proc_path = os.path.join(self.work_dir, 'blah', filename)
        chain.run(proc_path, image_data)
        im = Image.open(proc_path)
        self.assertEqual(im.size, (50, 50))

    def test_run_no_thumb(self):
        chain = FilterChain('thumbless', extension='png', no_thumb=True)
        im = self._process(chain, self.test_files[0])
        self.assertEqual(im.size, (512, 512))



"""
Test dogpile protection by using two reader processes to try to process the
same image chain. Ensure that the image chain is only processed once.
"""

from unittest import TestCase
from threading import Thread
from six.moves import queue

from pyramid import testing

from ..images.chain import FilterChain
from ..images.filters import Filter
from ..images.view import process_image

from . import utils


q = queue.Queue()


class RecordingFilter(Filter):
    def filter(self, im):
        q.put(1)
        return im


def run_process(name, original_ext):

    chain = FilterChain(
        'test-dogpile',
        filters=[
            RecordingFilter(),
        ],
        extension='jpg', width=50, height=50,
    )
    with testing.testConfig():
        process_image(utils.default_settings,
                      name, original_ext, chain)


def setup():
    utils.load_images()


class TestImageDogpileProtection(TestCase):

    def test_create(self):
        args = [
            ('smiley-jpeg-rgb', 'jpg'),
            ('smiley-jpeg-rgb', 'jpg'),
            ('smiley-png24-alpha', 'png'),
            ('smiley-png24-alpha', 'png'),
            ('smiley-png24-alpha', 'png'),
        ]
        threads = []
        for arg in args:
            t = Thread(target=run_process, args=arg)
            t.start()
            threads.append(t)
        for t in threads:
            t.join()
        # Make sure that the RecordingFilter was run exactly twice.
        self.assertEqual(q.qsize(), 2)



from __future__ import absolute_import, print_function, division

from webhelpers2.html.tags import HTML

from .asset import Asset


class SVGAsset(Asset):
    """
    Asset handler for SVG files.
    """
    extension = 'svg'

    def compile(self, key, theme, output_dir, minify=True):
        """
        Compile an SVG entry point.
        """
        entry_point = theme.static_url_to_filesystem_path(self.url_path)
        file_path = self.write_from_file(key, entry_point, self.url_path,
                                         output_dir)
        return file_path

    def tag(self, theme, url, **kw):
        return HTML.img(src=url, alt=kw.get('alt', ''), **kw)



from __future__ import absolute_import, print_function, division

from unittest import TestCase

from pyramid import testing
from pyramid_frontend.theme import Theme

from . import utils
from .example import base, foo, bar


class TestTheme(TestCase):
    def test_traverse_attributes(self):
        class A(Theme):
            key = 'a'
            template_dir = 'a-templates'

        class B(A):
            key = 'b'
            pass

        class C(B):
            key = 'c'
            template_dir = 'c-templates'

        self.assertEqual(list(C.traverse_attributes('template_dir')),
                         [('c', 'c-templates'),
                          ('b', 'a-templates'),
                          ('a', 'a-templates')])

    def test_repr(self):
        class BlahTheme(Theme):
            key = 'blah'

        self.assertIn('blah', repr(BlahTheme(())))


class TestThemeOpt(TestCase):
    def setUp(self):
        class A(Theme):
            key = 'a'

        class B(Theme):
            key = 'b'
            frobozz = 127

        class C(Theme):
            key = 'c'
            frobozz = 42

        self.a = A({})
        self.b = B({})
        self.c = C({})

    def test_opt_missing(self):
        with self.assertRaises(AttributeError):
            self.a.opt('frobozz')

    def test_opt_missing_with_default(self):
        self.assertEqual(self.a.opt('frobozz', 'somedefault'), 'somedefault')

    def test_opt_present(self):
        self.assertEqual(self.b.opt('frobozz'), 127)

    def test_opt_override(self):
        self.assertEqual(self.c.opt('frobozz'), 42)


class TestThemeStatic(TestCase):
    def setUp(self):
        self.base = base.BaseTheme({})
        self.foo = foo.FooTheme({})
        self.bar = bar.BarTheme({})

    def test_static_missing(self):
        with self.assertRaises(IOError):
            self.foo.static('images/nonexistent.jpg')

    def test_static_present(self):
        self.assertEqual(self.base.static('txt/sample.txt'),
                         '/_base/txt/sample.txt')

    def test_static_override(self):
        self.assertEqual(self.foo.static('txt/sample.txt'),
                         '/_foo/txt/sample.txt')


class TestIncludes(TestCase):
    def setUp(self):
        self.config = testing.setUp(settings=utils.default_settings)
        self.config.include('pyramid_frontend')

    def tearDown(self):
        testing.tearDown()

    def test_plain_include(self):
        self.config.add_theme(base.BaseTheme)
        settings = self.config.registry.settings
        included = settings['test_includes']
        self.assertEqual(included, ['base'])

    def test_subclass_include(self):
        self.config.add_theme(foo.FooTheme)
        settings = self.config.registry.settings
        included = settings['test_includes']
        self.assertEqual(included, ['base', 'foo'])

    def test_repeated_include(self):
        self.config.add_theme(bar.BarTheme)
        settings = self.config.registry.settings
        included = settings['test_includes']
        self.assertEqual(included, ['base'])



from __future__ import absolute_import, print_function, division

import pkg_resources
import os
import os.path
import glob
import shutil

from pyramid.config import Configurator

from pyramid_frontend.tests.example import base, foo, bar
from pyramid_frontend.images.files import check_and_save_image
from pyramid_frontend.images.chain import FilterChain

from .. import compile


samples_dir = pkg_resources.resource_filename('pyramid_frontend.tests', 'data')

test_ini_path = pkg_resources.resource_filename('pyramid_frontend.tests',
                                                'test.ini')

work_dir = '/tmp/pfe-test-workdir'


default_settings = {
    'pyramid_frontend.theme': 'foo',
    'pyramid_frontend.original_image_dir': os.path.join(work_dir, 'originals'),
    'pyramid_frontend.processed_image_dir': os.path.join(work_dir,
                                                         'processed'),
    'pyramid_frontend.compiled_asset_dir': os.path.join(work_dir, 'compiled'),
}


def noop_view(request):
    return {}


def bad_return_view(request):
    return 'not-a-dict'


def bad_template_view(request):
    return {}


def image_url_view(request):
    filter_key = request.params.get('filter_key', 'thumb')
    qualified = bool(request.params.get('qualified'))
    return request.image_url('smiley-jpeg-rgb', 'jpg', filter_key,
                             qualified=qualified)


def image_tag_view(request):
    filter_key = request.params.get('filter_key', 'thumb')
    return request.image_tag('smiley-jpeg-rgb', 'jpg', filter_key)


def image_original_path_view(request):
    return request.image_original_path('smiley-jpeg-rgb', 'jpg')


def js_tag_view(request):
    return request.asset_tag('main-js')


def css_tag_view(request):
    return request.asset_tag('main-less')


def svg_tag_view(request):
    return request.asset_tag('logo-svg')


def load_images(settings=default_settings):
    """
    Load images from test 'samples' directory to originals dir.
    """
    originals = settings['pyramid_frontend.original_image_dir']
    processed = settings['pyramid_frontend.processed_image_dir']
    if os.path.exists(originals):
        shutil.rmtree(originals)
    if os.path.exists(processed):
        shutil.rmtree(processed)
    os.makedirs(originals)
    os.makedirs(processed)

    for filename in glob.glob(os.path.join(samples_dir, '*.*')):
        f = open(filename, 'rb')
        name, raw_ext = os.path.basename(filename).rsplit('.', 1)
        if name != 'not-an-image':
            check_and_save_image(settings, name, f)


def make_app(settings=None, theme_strategy=None):
    base_settings = default_settings.copy()
    if settings:
        base_settings.update(settings)
    config = Configurator(settings=base_settings)

    config.include('pyramid_frontend')

    config.add_image_filter(FilterChain('thumb', width=200, height=200,
                                        crop=True))

    config.add_theme(base.BaseTheme)
    config.add_theme(foo.FooTheme)
    config.add_theme(bar.BarTheme)

    config.add_route('index', '/')
    config.add_route('article', '/article')

    config.add_route('bad-return', '/bad-return')
    config.add_route('bad-template', '/bad-template')

    config.add_route('image-url', '/image-url')
    config.add_route('image-tag', '/image-tag')
    config.add_route('image-original-path', '/image-original-path')

    config.add_route('js-tag', '/js-tag')
    config.add_route('css-tag', '/css-tag')
    config.add_route('svg-tag', '/svg-tag')

    config.add_route('text-template', '/text-template')

    config.add_view(noop_view, route_name='index', renderer='index.html')
    config.add_view(noop_view, route_name='article', renderer='article.html')

    config.add_view(bad_return_view, route_name='bad-return',
                    renderer='index.html')
    config.add_view(bad_template_view, route_name='bad-template',
                    renderer='bad.html')

    config.add_view(image_url_view, route_name='image-url', renderer='string')
    config.add_view(image_tag_view, route_name='image-tag', renderer='string')
    config.add_view(image_original_path_view, route_name='image-original-path',
                    renderer='string')

    config.add_view(js_tag_view, route_name='js-tag', renderer='string')
    config.add_view(css_tag_view, route_name='css-tag', renderer='string')
    config.add_view(svg_tag_view, route_name='svg-tag', renderer='string')

    config.add_view(noop_view, route_name='text-template',
                    renderer='text-template.txt')

    if theme_strategy:
        config.set_theme_strategy(theme_strategy)

    return config.make_wsgi_app()


def app_factory(global_config, **local_conf):
    global_config.update(local_conf)
    return make_app(global_config)


def compile_assets():
    args = ['pcompile', test_ini_path]
    return compile.main(args)


if __name__ == '__main__':
    from wsgiref.simple_server import make_server
    app = make_app()
    server = make_server('0.0.0.0', 8080, app)
    server.serve_forever()



from __future__ import absolute_import, print_function, division

import os
import os.path
import shutil

from .utils import work_dir


def setup():
    if os.path.exists(work_dir):
        shutil.rmtree(work_dir)
    os.makedirs(work_dir)



from __future__ import absolute_import, print_function, division

import os.path
import pkg_resources

from unittest import TestCase

from PIL import Image

from ..images import utils, filters

samples_dir = pkg_resources.resource_filename('pyramid_frontend.tests', 'data')


class TestUtils(TestCase):
    def setUp(self):
        self.im = Image.open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'))

    def test_bounding_box(self):
        self.assertEqual(utils.bounding_box(self.im), (8, 16, 496, 504))

    def test_image_entropy(self):
        entropy = utils.image_entropy(self.im)
        self.assertAlmostEqual(entropy, 5.4, places=1)

    def test_is_larger(self):
        self.assertTrue(utils.is_larger(self.im, (200, 200)))
        self.assertFalse(utils.is_larger(self.im, (600, 600)))
        self.assertFalse(utils.is_larger(self.im, (512, 512)))

    def test_pad_image(self):
        padded = utils.pad_image(self.im, (600, 600))
        self.assertEqual(padded.size, (600, 600))

    def test_colors_differ(self):
        self.assertTrue(utils.colors_differ((127, 243, 100),
                                            (125, 243, 95),
                                            tolerance=5))
        self.assertFalse(utils.colors_differ((127, 243, 100),
                                             (125, 243, 95),
                                             tolerance=20))

    def test_crop_entropy_same(self):
        im = utils.crop_entropy(self.im, (300, 300))
        self.assertEqual(im.getpixel((150, 150)), (206, 205, 1))

    def test_crop_entropy_wide(self):
        im = utils.crop_entropy(self.im, (300, 100))
        self.assertEqual(im.getpixel((150, 50)), (202, 200, 1))

    def test_crop_entropy_tall(self):
        im = utils.crop_entropy(self.im, (100, 300))
        self.assertIn(im.getpixel((50, 150)), [(208, 208, 0),
                                               (209, 208, 0)])

    def test_is_white_background(self):
        self.assertTrue(utils.is_white_background(self.im))

        cropped = self.im.crop((0, 0, 256, 256))
        self.assertFalse(utils.is_white_background(cropped))
        cropped = self.im.crop((0, 256, 256, 512))
        self.assertFalse(utils.is_white_background(cropped))
        cropped = self.im.crop((0, 0, 256, 512))
        self.assertFalse(utils.is_white_background(cropped))
        cropped = self.im.crop((256, 0, 512, 512))
        self.assertFalse(utils.is_white_background(cropped))


class TestFilters(TestCase):
    def test_png_save_rgb(self):
        saver = filters.PNGSaver(palette=True)
        im = Image.new("RGB", (25, 25), "red")
        f = saver(im)

        nm = Image.open(f)
        self.assertEqual(nm.mode, "P")
        nm = nm.convert("RGBA")
        self.assertEqual(nm.getpixel((8, 8)), (255, 0, 0, 255))

    def test_png_save_palette(self):
        saver = filters.PNGSaver(palette=True)
        im = Image.new("P", (25, 25))
        im.putpalette([0] * 768)
        im.paste(23, (0, 0, 25, 25))
        f = saver(im)

        nm = Image.open(f)
        self.assertEqual(nm.mode, "P")
        self.assertEqual(nm.getpixel((8, 8)), 23)



from __future__ import absolute_import, print_function, division

import os
import shutil
import hashlib

from datetime import datetime

from PIL import Image


filter_sep = '_'


def prefix_for_name(name):
    """
    Return the 4-char hash prefix to use for this image name (prevents having
    too many images in the same directory).
    """
    return hashlib.md5(name.encode('utf-8')).hexdigest()[:4]


def get_url_prefix(settings):
    return settings.get('pyramid_frontend.image_url_prefix', '/img').\
        rstrip('/')


def save_image(settings, name, original_ext, f):
    prefix = prefix_for_name(name)
    ensure_dirs(settings, prefix)
    save_locally(original_path(settings, name, original_ext), f)


def save_locally(path, f):
    diskf = open(path, 'wb')
    f.seek(0)
    shutil.copyfileobj(f, diskf)
    diskf.close()


def check(f):
    """
    Given a file object, check to see if the contents is a valid image. If so,
    return the file format. Otherwise, raise exceptions.
    """
    im = Image.open(f)
    # Raises exceptions if the file is broken in any way.
    im.verify()
    return (im.format, im.mode)


def save_to_error_dir(settings, name, f):
    """
    Save a questionable image (could not be verified by PIL) to a penalty box
    for investigation.
    """
    prefix = datetime.utcnow().strftime('%m-%d-%Y.%H:%M:%S')
    dirpath = os.path.join(settings['pyramid_frontend.error_dir'], prefix)
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)

    filepath = os.path.join(dirpath, name)
    save_locally(filepath, f)


def ensure_dirs(settings, prefix):
    dirpath = os.path.join(settings['pyramid_frontend.original_image_dir'],
                           prefix)
    if not os.path.exists(dirpath):
        os.makedirs(dirpath)


def original_path(settings, name, original_ext):
    dir = settings['pyramid_frontend.original_image_dir']
    return os.path.join(dir,
                        prefix_for_name(name),
                        '%s.%s' % (name, original_ext))


def processed_path(settings, name, original_ext, chain):
    dir = settings['pyramid_frontend.processed_image_dir']
    return os.path.join(
        dir,
        prefix_for_name(name),
        chain.basename(name, original_ext))


def check_and_save_image(settings, name, f):
    """
    Save an image to the local ``pyramid_frontend`` image originals directory,
    using the supplied base name and file-like object.

    The extension is chosen and normalized based on file format, and returned.
    It will always be three characters.

    If the image is not valid (cannot be loaded as a PIL image), it is saved to
    the error directory, and the exception raised by PIL is re-raised.
    """
    try:
        im = Image.open(f)
        format = im.format
        size = im.size
    except IOError:
        save_to_error_dir(settings, name, f)
        raise

    possible_extensions = {
        'JPEG': 'jpg',
        'PNG': 'png',
        'TIFF': 'tif',
        'GIF': 'gif',
    }
    original_ext = possible_extensions[format]
    save_image(settings, name, original_ext, f)
    f.close()
    return dict(ext=original_ext, size=size)



from __future__ import absolute_import, print_function, division

import os.path
from unittest import TestCase

from mako.exceptions import TopLevelLookupException
from mako.runtime import Context
from mako.util import FastEncodingBuffer

from ..templating.lookup import SuperTemplateLookup

from .example import base, foo, quux


class TestThemeLookup(TestCase):
    settings = {
        'pyramid_frontend.compiled_asset_dir': '.'
    }

    def test_render_foo_index(self):
        theme = foo.FooTheme(self.settings)
        templ = theme.lookup.get_template('index.html')
        s = templ.render()
        self.assertIn(b'base theme base.html', s)
        self.assertIn(b'foo theme base.html', s)
        self.assertIn(b'base theme index.html', s)
        self.assertNotIn(b'bar', s)

        self.assertIn(b'foo theme blocks header', s)

    def test_render_foo_article(self):
        theme = foo.FooTheme(self.settings)
        templ = theme.lookup.get_template('article.html')
        s = templ.render()
        self.assertIn(b'base theme base.html', s)
        self.assertIn(b'foo theme base.html', s)
        self.assertIn(b'foo theme article.html', s)
        self.assertNotIn(b'bar', s)

        self.assertIn(b'foo theme blocks header', s)

    def test_render_base_index(self):
        theme = base.BaseTheme(self.settings)
        templ = theme.lookup.get_template('index.html')
        s = templ.render()
        self.assertIn(b'base theme base.html', s)
        self.assertIn(b'base theme index.html', s)
        self.assertNotIn(b'foo', s)
        self.assertNotIn(b'bar', s)

        self.assertIn(b'base theme blocks header', s)

    def test_render_base_article(self):
        theme = base.BaseTheme(self.settings)
        templ = theme.lookup.get_template('article.html')
        s = templ.render()
        self.assertIn(b'base theme base.html', s)
        self.assertIn(b'base theme article.html', s)
        self.assertNotIn(b'foo', s)

        self.assertIn(b'base theme blocks header', s)

    def test_render_quux_article(self):
        theme = quux.QuuxTheme(self.settings)
        templ = theme.lookup.get_template('article.html')
        s = templ.render()
        self.assertIn(b'base theme base.html', s)
        self.assertIn(b'foo theme base.html', s)
        self.assertIn(b'foo theme article.html', s)

        self.assertIn(b'quux theme blocks header', s)

    def test_render_missing_template(self):
        theme = foo.FooTheme(self.settings)
        with self.assertRaises(TopLevelLookupException):
            theme.lookup.get_template('nonexistent-template.html')

    def test_find_dir_for(self):
        theme = foo.FooTheme(self.settings)
        with self.assertRaises(ValueError):
            theme.lookup.find_dir_for_path('/some/other/path.html')

    def test_render_text(self):
        theme = foo.FooTheme(self.settings)
        templ = theme.lookup_nofilters.get_template('article.txt')
        s = templ.render()
        self.assertIn(b'foo theme article.txt', s)
        self.assertIn(b'this should not be escaped: <>', s)

    def test_render_text_unicode(self):
        theme = foo.FooTheme(self.settings)
        templ = theme.lookup_nofilters.get_template('unicode.txt')
        s = templ.render()
        self.assertIn(b'foo theme unicode.txt', s)
        self.assertIn(u'this is unicode: \u2603'.encode('utf8'), s)

    def test_get_def_base(self):
        theme = base.BaseTheme(self.settings)
        templ = theme.lookup.get_template('common/blocks.html')
        buf = FastEncodingBuffer(as_unicode=True)
        ctx = Context(buf)
        templ.get_def('hello').render_context(ctx)

        s = buf.getvalue()
        self.assertIn('base theme blocks hello', s)

    def test_get_def_by_file(self):
        theme = quux.QuuxTheme(self.settings)

        filename = os.path.join(theme.template_dirs[1], 'common/blocks.html')
        uri = theme.lookup.filename_to_uri(filename)

        templ = theme.lookup.get_template(uri)

        buf = FastEncodingBuffer(as_unicode=True)
        ctx = Context(buf)
        templ.get_def('hello').render_context(ctx)

        s = buf.getvalue()
        self.assertIn('foo theme blocks hello', s)


class TestLookupOptions(TestCase):
    settings = {
        'pyramid_frontend.compiled_asset_dir': '.'
    }

    def test_filesystem_checks(self):
        theme = foo.FooTheme(self.settings)
        dirs = theme.lookup.directories
        lookup = SuperTemplateLookup(directories=dirs,
                                     filesystem_checks=True)
        templ = lookup.get_template('index.html')
        s = templ.render()
        self.assertIn('foo theme', s)



from __future__ import absolute_import, print_function, division

import logging

import argparse
import sys

from pyramid.paster import bootstrap


def compile(registry, minify=True, only_theme=None):
    """
    Compile static assets for all themes which are registered in ``registry``.
    """
    log = logging.getLogger('pyramid_frontend')
    settings = registry.settings
    theme_registry = settings['pyramid_frontend.theme_registry']
    themes = theme_registry.values()
    count = len(themes)
    for ii, theme in enumerate(themes):
        if only_theme and (theme.key != only_theme):
            log.warn("%d / %d - Skipping theme: %s", ii + 1, count, theme.key)
        else:
            log.warn("%d / %d - Compiling theme: %s", ii + 1, count, theme.key)
            theme.compile(minify=minify)


class ConsoleHandler(logging.StreamHandler):
    """
    A subclass of StreamHandler which behaves in the same way, but colorizes
    the log level before formatting it.
    """

    colors = {'CRITICAL': '[1;31m',
              'ERROR':    '[1;31m',
              'WARNING':  '[1;33m',
              'INFO':     '[1;32m',
              'DEBUG':    '[1;37m',
              None:    '[0m'}

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record, but
        before the formatter is applied the loglevel is colored.
        """

        record.colored_levelname = (self.colors[record.levelname] +
                                    record.levelname +
                                    self.colors[None])
        sys.stdout.write(self.format(record) + '\n')


def configure_logging(verbosity):
    """
    Configure logging for use with the asset compilation command.
    """
    levels = [logging.CRITICAL,
              logging.ERROR,
              logging.WARNING,
              logging.INFO,
              logging.DEBUG]

    ch = ConsoleHandler()
    ch.setLevel(levels[verbosity])

    formatter = logging.Formatter(
        '%(asctime)s %(colored_levelname)s %(name)s - %(message)s')
    ch.setFormatter(formatter)

    log = logging.getLogger('pyramid_frontend')
    log.setLevel(logging.DEBUG)
    log.addHandler(ch)
    return log


def main(args=sys.argv):
    """
    Main entry point for the executable which compiles assets.
    """
    parser = argparse.ArgumentParser(description='Compile static assets.')
    parser.add_argument('--no-minify', action='store_true', default=False)
    parser.add_argument('-v', '--verbose', action='count', default=2)
    parser.add_argument('-t', '--theme', default=None)
    parser.add_argument('config_uri')

    options = parser.parse_args(args[1:])

    env = bootstrap(options.config_uri)
    configure_logging(options.verbose)
    registry = env['registry']
    compile(registry, minify=(not options.no_minify), only_theme=options.theme)
    return 0



from __future__ import absolute_import, print_function, division

import os.path
import subprocess
from mock import patch
from unittest import TestCase
from six import StringIO

from .. import compile, cmd
from ..assets.less import LessAsset
from ..assets.requirejs import RequireJSAsset, js_preamble
from ..assets.svg import SVGAsset

from . import utils
from .example import foo


class TestCompileCommand(TestCase):

    def test_pcompile_usage(self):
        args = [
            'pcompile',
        ]
        buf = StringIO()
        with patch('sys.stderr', buf):
            with self.assertRaises(SystemExit) as cm:
                compile.main(args)
            exit_exception = cm.exception
            self.assertEqual(exit_exception.code, 2)
        self.assertIn('config_uri', buf.getvalue())

    def test_pcompile(self):
        args = ['pcompile', utils.test_ini_path]
        retcode = compile.main(args)
        self.assertEqual(retcode, 0)

    def test_pcompile_no_minify(self):
        args = ['pcompile', utils.test_ini_path, '--no-minify']
        retcode = compile.main(args)
        self.assertEqual(retcode, 0)
        # XXX Try to assett hat this makes longer assets or something.


class TestAsset(TestCase):
    def setUp(self):
        self.theme = foo.FooTheme({})
        self.output_dir = os.path.join(utils.work_dir, 'compile-tests')

    def test_bad_shell(self):
        buf = StringIO()
        with patch('sys.stdout', buf):
            with self.assertRaises(subprocess.CalledProcessError):
                argv = [
                    'false',
                ]
                cmd.run(argv)
        # XXX Try to test that this actually prints the stdout output of a
        # failed comamnd.

    def test_less_compile(self):
        asset = LessAsset('/_foo/css/main.less')
        path = asset.compile(key='main-less',
                             theme=self.theme,
                             output_dir=self.output_dir)
        f = open(path, 'rb')
        buf_minified = f.read()
        self.assertGreater(len(buf_minified), 0)

        # Test that minified version is smaller than non-minified.
        asset = LessAsset('/_foo/css/main.less')
        path = asset.compile(key='main-less',
                             theme=self.theme,
                             output_dir=self.output_dir,
                             minify=False)
        f = open(path, 'rb')
        buf_unminified = f.read()

        self.assertLess(len(buf_minified), len(buf_unminified))

    def test_requirejs_compile(self):
        asset = RequireJSAsset('/_foo/js/main.js')
        path = asset.compile(key='main-js',
                             theme=self.theme,
                             output_dir=self.output_dir)
        f = open(path, 'rb')
        buf_minified = f.read()
        self.assertGreater(len(buf_minified), 0)

        # Test that minified version is smaller than non-minified.
        preamble_size = len(js_preamble)
        asset = RequireJSAsset('/_foo/js/main.js')
        path = asset.compile(key='main-js',
                             theme=self.theme,
                             output_dir=self.output_dir,
                             minify=False)
        f = open(path, 'rb')
        buf_unminified = f.read()

        self.assertLess(len(buf_minified) - preamble_size, len(buf_unminified))

    def test_svg_compile(self):
        asset = SVGAsset('/_foo/images/logo.svg')
        path = asset.compile(key='logo-svg',
                             theme=self.theme,
                             output_dir=self.output_dir)
        f = open(path, 'r')
        buf = f.read()
        self.assertGreater(len(buf), 0)
        self.assertIn('svg', buf)



from __future__ import absolute_import, print_function, division

import logging

import os
import io

from webhelpers2.html.tags import HTML, literal

from .. import cmd
from .asset import Asset

log = logging.getLogger(__name__)


js_preamble = '''\
if (typeof console === 'undefined') {
  console = {
    log: function () {},
    debug: function () {}
  }
}
'''


def render_js_paths(theme):
    """
    Return a script tag for use client-side which sets up require.js paths for
    all theme directories in use by the supplied theme.
    """
    cls = theme.__class__
    keys = []
    while hasattr(cls, 'key'):
        keys.append(cls.key)
        cls = cls.__bases__[0]
    lines = ["require.paths.%s = '/_%s/js';" % (key, key)
             for key in keys]
    return ''.join(['<script>'] + lines + ['</script>'])


class RequireJSAsset(Asset):
    """
    Asset handler for javascript loaded by require.js.

    Currently assumes r.js is on $PATH.
    """
    extension = 'js'

    def __init__(self, url_path,
                 require_config_path='/_pfe/require_config.js',
                 require_path='/_pfe/require.js',
                 require_base_url='/_pfe/'):
        self.url_path = url_path
        self.require_config_path = require_config_path
        self.require_path = require_path
        self.require_base_url = require_base_url

    def compile(self, key, theme, output_dir, minify=True):
        main_config = theme.static_url_to_filesystem_path(
            self.require_config_path)
        base_url = theme.static_url_to_filesystem_path(
            self.require_base_url)

        log.debug("base_url: %r", base_url)
        log.debug("main_config: %r", main_config)

        # Path to main module relative to baseUrl (can't be absolute)
        main_js_file = theme.static_url_to_filesystem_path(self.url_path)
        main = os.path.relpath(main_js_file, base_url)
        main = os.path.splitext(main)[0]  # Strip .js

        almond_path = \
            theme.static_url_to_filesystem_path('/_pfe/almond.js')
        almond_path = os.path.relpath(almond_path, base_url)
        almond_path = os.path.splitext(almond_path)[0]

        log.debug("main: %r", main)

        args = [
            'r.js', '-o',
            'baseUrl={0}'.format(base_url),
            'mainConfigFile={0}'.format(main_config),
            'name={0}'.format(main),
            'paths.requireLib={0}'.format(almond_path),
            'include=requireLib',
        ]

        if not minify:
            args.append('optimize=none')

        # Add RequireJS paths for theme
        for dir_ref, dir in theme.keyed_static_dirs:
            dir = os.path.join(dir, 'js')
            log.debug("path _%s -> %s", key, dir)
            args.append('paths.{}={}'.format(dir_ref, dir))

        with self.tempfile() as (f, temp_name):
            args.append('out={0}'.format(temp_name))
            cmd.run(args)

            with io.open(temp_name, encoding='utf8') as f:
                contents = f.read()

            contents = '\n'.join([js_preamble, contents])
            file_path = self.write(key, contents, self.url_path, output_dir)

        return file_path

    def tag_development(self, theme, url):
        """
        Return an HTML fragment to use a require.js entry point in development.
        """
        return ''.join([
            HTML.script(literal(js_preamble)),
            HTML.script(src=self.require_config_path),
            render_js_paths(theme),
            HTML.script(src=self.require_path),
            HTML.script(src=url),
        ])

    def tag_production(self, theme, url):
        """
        Return an HTML fragment to use a require.js entry point in production.
        """
        return HTML.script(src=url)



from __future__ import print_function

import os
from setuptools import setup, find_packages
from distutils.command.build import build as _build


executables = [
    'jpegoptim',
    'pngcrush',
    'optipng',
    'convert',  # From GraphicsMagick or ImageMagick.
    'lessc',
    'r.js',
    'postcss',
]


def which(exe):
    for path in os.environ['PATH'].split(os.pathsep):
        fpath = os.path.join(path, exe)
        if os.path.exists(fpath) and os.access(fpath, os.X_OK):
            return fpath


class build(_build):
    def finalize_options(self):
        _build.finalize_options(self)
        for exe in executables:
            print('checking for %r...' % exe)
            fpath = which(exe)
            if not fpath:
                print('WARNING: Missing %r executable!' % exe)
            else:
                print(fpath)


setup(name='pyramid_frontend',
      cmdclass={'build': build},
      version='0.6.5',
      description='Themes, image filtering, and frontend asset handling.',
      long_description='',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Framework :: Pyramid',
          'Intended Audience :: Developers',
          'Topic :: Internet :: WWW/HTTP',
          'Topic :: Multimedia :: Graphics',
      ],
      keywords='pyramid themes frontend assets',
      url='http://github.com/cartlogic/pyramid_frontend',
      author='Scott Torborg',
      author_email='scott@cartlogic.com',
      install_requires=[
          'Pyramid>=1.4.5',
          'Pillow>=2.1.0',      # Provides PIL
          'Mako>=0.9.0',
          'WebHelpers2>=2.0b5',
          'six>=1.5.2',
          'lockfile>=0.9.1',
      ],
      license='MIT',
      packages=find_packages(),
      test_suite='nose.collector',
      tests_require=['nose'],
      include_package_data=True,
      zip_safe=False,
      entry_points="""\
      [console_scripts]
      pcompile = pyramid_frontend.compile:main
      """)



from __future__ import absolute_import, print_function, division

import os.path
import pkg_resources

from unittest import TestCase

from ..images import files

from . import utils

samples_dir = pkg_resources.resource_filename('pyramid_frontend.tests', 'data')


class TestFiles(TestCase):
    def test_check_bad(self):
        f = open(os.path.join(samples_dir, 'not-an-image.png'), 'rb')
        with self.assertRaises(IOError):
            files.check(f)

    def test_check_good(self):
        f = open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'), 'rb')
        files.check(f)

    def test_save_to_error_dir(self):
        name = 'not-an-image.png'
        f = open(os.path.join(samples_dir, name), 'rb')
        error_dir = os.path.join(utils.work_dir, 'errors')
        settings = {
            'pyramid_frontend.error_dir': error_dir,
        }
        files.save_to_error_dir(settings, name, f)
        # Saving same file twice should work
        files.save_to_error_dir(settings, '2.png', f)

    def test_check_and_save_image_bad(self):
        f = open(os.path.join(samples_dir, 'not-an-image.png'), 'rb')
        error_dir = os.path.join(utils.work_dir, 'errors')
        settings = {
            'pyramid_frontend.error_dir': error_dir,
        }
        with self.assertRaises(IOError):
            files.check_and_save_image(settings, 'not-an-image', f)

    def test_check_and_save_image_good(self):
        f = open(os.path.join(samples_dir, 'smiley-jpeg-rgb.jpg'), 'rb')
        originals_dir = os.path.join(utils.work_dir, 'originals')
        settings = {
            'pyramid_frontend.original_image_dir': originals_dir,
        }
        info = files.check_and_save_image(settings, 'smiley-jpeg-rgb', f)
        self.assertEqual(info['ext'], 'jpg')
        self.assertEqual(info['size'], (512, 512))



from __future__ import absolute_import, print_function, division

from unittest import TestCase

from pyramid import testing
from ..images.chain import FilterChain


class TestConfig(TestCase):
    settings = {
        'pyramid_frontend.compiled_asset_dir': '.'
    }

    def setUp(self):
        self.config = testing.setUp(settings=self.settings)
        self.config.include('pyramid_frontend')

    def tearDown(self):
        testing.tearDown()

    def test_directives_get_registered(self):
        self.assertTrue(hasattr(self.config, 'add_image_filter'))
        self.assertTrue(hasattr(self.config, 'add_theme'))

    def test_register_filter_chain_twice(self):
        chain = FilterChain('blah', width=200, height=200,
                            crop=True, extension='png')
        self.config.add_image_filter(chain, None)
        self.config.add_image_filter(chain, None)

    def test_register_conflicting_filters(self):
        chain = FilterChain('blah', width=200, height=200,
                            crop=True, extension='png')
        self.config.add_image_filter(chain, None)
        chain = FilterChain('blah', width=200, height=200,
                            crop=True, extension='jpg', quality=85)
        with self.assertRaises(ValueError):
            self.config.add_image_filter(chain, None)



from __future__ import absolute_import, print_function, division

import logging

import os
import tempfile
import io

from contextlib import contextmanager
from hashlib import sha1

import six

log = logging.getLogger(__name__)


class Asset(object):
    """
    Generic superclass for other asset handler classes to inherit from.
    """

    def __init__(self, url_path):
        self.url_path = url_path

    def write(self, key, contents, entry_point, output_dir):
        """
        Write the compiled result for a particular entry point to the
        appropriate file and map file.

        The ``contents`` field should always be a unicode type (str on 3.x).
        """
        assert isinstance(contents, six.text_type)
        log.debug('Write - key: %r, entry_point: %r', key, entry_point)
        hash = sha1(contents.encode('utf-8')).hexdigest()

        name = os.path.basename(entry_point)
        file_name = u'{name}-{hash}.{ext}'.format(
            name=name, hash=hash, ext=self.extension)
        file_path = os.path.join(output_dir, file_name)

        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        log.debug('Writing to %s ...', file_path)
        with io.open(file_path, 'w', encoding='utf8') as f:
            f.write(contents)

        map_path = os.path.join(output_dir, key + '.map')
        log.debug('Writing map file to %s ...', map_path)
        with io.open(map_path, 'w') as f:
            f.write(file_name)

        return file_path

    def write_from_file(self, key, file_name, entry_point, output_dir):
        """
        Like ``write()``, but writes from a source file instead of a buffer
        variable.
        """
        with io.open(file_name, encoding='utf8') as f:
            contents = f.read()
        return self.write(key, contents, entry_point, output_dir)

    @contextmanager
    def tempfile(self, *args, **kwargs):
        """
        A context manager helper to generate a temporary file for quick use.
        """
        f = tempfile.NamedTemporaryFile()
        try:
            yield f, f.name
        finally:
            f.close()

    def tag(self, theme, url, production=True):
        if production:
            return self.tag_production(theme, url)
        else:
            return self.tag_development(theme, url)
